import java.util.Collection;
import java.util.Iterator;
import java.util.NoSuchElementException;
import java.util.Set;

/**
 * Classe degli insiemi basati su array.
 *
 * @author  Federico Bergenti
 * @version 1.00 10/06/03
 */
public class ArraySet implements Set, Cloneable {
    /**
     * Ritorna la cardinalita' dell'insieme.
     *
     * @return La cardinalita' dell'insieme.
     */
    public int size() {
        return size;
    }

    /**
     * Verifica se l'insieme e' vuoto.
     *
     * @return Vero se l'insieme e' vuoto.
     */
    public boolean isEmpty() {
        return size == 0;
    }

    /**
     * Svuota l'insieme.
     */
    public void clear() {
        size       = 0;
        bufferSize = initialSize;
        buffer     = new Object[bufferSize];
    }

    /**
     * Verifica se un oggetto e' contenuto nell'insieme.
     *
     * @param  o L'oggetto da cercare.
     * @return Una indicazione della presenza di <code>o</code> nell'insieme.
     */
    public boolean contains(Object o) {
        for(int i = 0; i < size; i++)
            if(equals(buffer[i], o)) return true;

        return false;
    }

    /**
     * Verifica se gli oggetti della <code>Collection</code> passata come parametro sono contenuti nell'insieme.
     *
     * @param  c La <code>Collection</code> da verificare.
     * @return Una indicazione della presenza di <code>o</code> nell'insieme.
     */
    public boolean containsAll(Collection c) {
        if(c == null) throw new NullPointerException();

        for(Iterator i = c.iterator(); i.hasNext(); )
            if(!contains(i.next())) return false;

        return true;
    }

    /**
     * Aggiunge un oggetto all'insieme.
     *
     * @param o L'oggetto da aggiungere.
     * @return Una indicazione del cambiamento dell'insieme a causa dell'operazione.
     */
    public boolean add(Object o) {
        if(contains(o)) return false;

        trimBuffer();

        buffer[size++] = o;

        return true;
    }

    /**
     * Aggiunge gli oggetti della <code>Collection</code> passata come parametro all'insieme.
     *
     * @param  c La <code>Collection</code> i cui oggetti vanno aggiunti all'insieme.
     * @return Una indicazione del cambiamento dell'insieme a causa dell'operazione.
     */
    public boolean addAll(Collection c) {
        if(c == null) throw new NullPointerException();

        boolean changed = false;

        for(Iterator i = c.iterator(); i.hasNext(); )
            if(add(i.next())) changed = true;

        return changed;
    }

    /**
     * Rimuove un oggetto dall'insieme.
     *
     * @return Un iteratore sull'insieme.
     */
    public boolean remove(Object o) {
        boolean found = false;

        for(int i = 0; i < size-1; i++) {
            if(found) {
                buffer[i] = buffer[i+1];
            } else {
                if(equals(buffer[i], o)) { buffer[i] = buffer[i+1]; found = true; }
            }
        }

        if(!found && equals(buffer[size-1], o)) found = true;

        if(found) { size--; trimBuffer(); }

        return found;
    }

    /**
     * Rimuove gli oggetti della <code>Collection</code> passata come parametro all'insieme.
     *
     * @param  c La <code>Collection</code> i cui oggetti vanno rimossi dall'insieme.
     * @return Una indicazione del cambiamento dell'insieme a causa dell'operazione.
     */
    public boolean removeAll(Collection c) {
        if(c == null) throw new NullPointerException();

        boolean changed = false;

        for(Iterator i = c.iterator(); i.hasNext(); )
            if(remove(i.next())) changed = true;

        return changed;
    }

    /**
     * Mantiene nell'insieme solo gli oggetti della <code>Collection</code> passata come parametro.
     *
     * @param  c La <code>Collection</code> i cui oggetti vanno rimossi dall'insieme.
     * @return Una indicazione del cambiamento dell'insieme a causa dell'operazione.
     */
    public boolean retainAll(Collection c) {
        if(c == null) throw new NullPointerException();

        boolean changed = false;

        for(int i = 0; i < size; i++)
            if(!c.contains(buffer[size])) { 
                remove(buffer[size]); changed = true; 
            }

        return changed;
    }

    /**
     * Ritorna un iteratore sull'insieme.
     *
     * @return Un iteratore sull'insieme.
     */
    public Iterator iterator() {
        return new ArraySetIterator(this);
    }

    /**
     * Crea una deep copy dell'insieme.
     *
     * @return Una deep copy dell'insieme.
     */
    public Object clone() {
        ArraySet result = new ArraySet(initialSize);

        for(int i = 0; i < size; i++) result.add(buffer[i]);

        return result;
    }

    /**
     * Converte l'insieme in un array.
     *
     * @return Un nuovo array contenente l'insieme.
     */
    public Object[] toArray() {
        Object[] result = new Object[size];

        for(int i = 0; i < size; i++) result[i] = buffer[i];

        return result;
    }

    /**
     * Converte l'insieme in un array (non implementato, per semplicita').
     *
     * @param target L'array da riempire con gli oggetti contenuti nell'insieme.
     * @return Un nuovo array contenente l'insieme.
     */
    public Object[] toArray(Object[] target) {
        throw new UnsupportedOperationException();
    }

    /**
     * Costruttore per un insieme basato su un array.
     *
     * @param is Dimensione iniziale dell'array a cui si appoggia l'insieme.
     */
    public ArraySet(int is) {
        initialSize = is;

        clear();
    }

    /**
     * Confronta due oggetti (anche nulli).
     *
     * @param o1 Il primo oggetto.
     * @param o2 Il secondo oggetto.
     * @return Vero se i due oggetti sono uguali.
     */
    private boolean equals(Object o1, Object o2) {
        if(o1 != null) {
            if(o1.equals(o2)) return true;
        } else {
            if(o2 == null) return true;
        }

        return false;
    }

    /**
     * Crea nuove celle nell'array o ne cancella in caso di necessita'.
     */
    private void trimBuffer() {
        int      newBufferSize = bufferSize;

        if(size < bufferSize*0.25) newBufferSize = (int)(bufferSize*0.75);
        if(size > bufferSize*0.75) newBufferSize = (int)(bufferSize*1.25);

        if(newBufferSize == bufferSize) return;

        Object[] result = new Object[newBufferSize];

        for(int i = 0; i < size; i++) result[i] = buffer[i];

        bufferSize = newBufferSize;
        buffer     = result;
    }

    /**
     * Dimensione attuale dell'array.
     */
    private int bufferSize = 0;

    /**
     * Dimensione iniziale dell'array.
     */
    private int initialSize = 0;

    /**
     * Cardinalita' attuale dell'array.
     */
    private int size = 0;

    /**
     * L'array su cui si appoggia l'insieme.
     */
    private Object[] buffer = null;

    /**
     * Classe degli iteratori per insiemi basati su array.
     */
    private class ArraySetIterator implements Iterator {
        /**
         * Ritorna l'oggetto corrente nell'attraversamento.
         *
         * @return L'oggetto corrente nell'attraversamento.
         */
        public Object next() {
            if(current == collection.size) throw new NoSuchElementException();

            return collection.buffer[current++];
        }

        /**
         * Ritorna <code>true</code> se ci sono ancora oggetti nell'attraversamento.
         *
         * @return Vero se ci sono ancora oggetti nell'attraversamento.
         */
        public boolean hasNext() {
            return current < collection.size;
        }

        /**
         * Rimuove l'oggetto corrente nell'attraversamento (non implementato).
         *
         * @return Vero se ci sono ancora oggetti nell'attraversamento.
         */
        public void remove() {
            throw new UnsupportedOperationException();
        }

        /**
         * Costruisce un iteratore su un insieme basato su un array.
         *
         * @param arraySet L'insieme da attraversare.
         */
        private ArraySetIterator(ArraySet arraySet) {
            collection = arraySet;
        }

        /**
         * Posizione corrente nella scansione.
         */
        private int current = 0;

        /**
         * L'insieme da attraversare.
         */
        private ArraySet collection = null;
    }
}
