package bancomat;

import bancomat.test.*;

import java.io.*;

/**
 * Classe dei controller del sistema Bancomat.
 *
 * @author  Federico Bergenti
 * @version 1.00 10/06/03
 */
public class Controller {
    /**
     * Costruttore del controller.
     */
    public Controller(boolean testMode) {
        test = testMode;

        if(test) {
            // crea un oggetto Bank ed un oggetto GUI di test
            gui =  new TestGUI(this);
            bank = new TestBank(this);
        } else {
            // crea un oggetto Bank ed un oggetto GUI reali
            throw new RuntimeException("Il sistema puo' essere lanciato solo in modalita' test");
        }
    }

    /**
     * Attiva il sistema.
     */
    public void run() {
        try { 
            logFile = new PrintStream(new FileOutputStream("bancomat.log"));
        } catch(Throwable t) {
            logFile = System.err;

            logError("Impossibile aprire il file di log, verra' utilizzato lo standard output");
        }

        state = INITIAL_STATE;

        logDebug("Sistema entrato nello stato INITIAL_STATE");

        bank.run(); gui.run();

        logDebug("GUI e banca attivate");

        logInfo("Sistema partito");
    }

    /**
     * E' stato introdotto un nuovo numero di tessera.
     */
    public void cardNumberEntered(String number) {
        if(state != INITIAL_STATE) { error("Errore interno del sistema"); return; }

        if(number == null || number.length() == 0) {
            error("Impossibile leggere il numero di carta"); return;
        }

        card = new Card(number);

        state = REQUEST_PIN_STATE; 

        logDebug("Sistema entrato nello stato REQUEST_PIN_STATE");

        gui.showRequestPINScreen();        
    }

    /**
     * E' stato introdotto un nuovo PIN.
     */
    public void pinEntered(String number) {
        if(state != REQUEST_PIN_STATE) { error("Errore interno del sistema"); return; }

        if(number == null || number.length() == 0) {
            error("Impossibile validare un PIN nullo"); return;
        }

        state = VALIDATE_ACCOUNT_STATE;

        logDebug("Sistema entrato nello stato VALIDATE_ACCOUNT_STATE");

        pin = new PIN(number);

        try {
            // valida la coppia numero di carta/PIN presso la banca e crea l'oggetto
            // conto corrente
            bank.validate(card.getNumber(), number);
        } catch(InvalidPINException ipe) { error("PIN errato"); }
    }

    /**
     * La banca ha validato con successo la coppia numero di carta/PIN.
     */
    public void accountFound(Account a) {
        if(state != VALIDATE_ACCOUNT_STATE) { error("Errore interno del sistema"); return; }

        account = a; 
        user    = new User(card, pin, account);

        logInfo("Cliente con numero di carta " + card.getNumber() + " entrato nel sistema");

        state = MAIN_MENU_STATE; 

        logDebug("Sistema entrato nello stato MAIN_MENU_STATE");

        gui.showMainMenuScreen();
    }

    /**
     * E' stato scelto di prelevare.
     */
    public void getMoneyChosen() {
        if(state != MAIN_MENU_STATE) { error("Errore interno del sistema"); return; }

        state = ENTER_AMOUNT_STATE;

        logDebug("Sistema entrato nello stato ENTER_AMOUNT_STATE");

        gui.showRequestMoneyScreen();
    }

    /**
     * E' stata introdotta la somma da prelevare.
     */
    public void amountEntered(String number) {
        if(state != ENTER_AMOUNT_STATE) { error("Errore interno del sistema"); return; }

        if(number == null || number.length() == 0) {
            error("Impossibile prelevare una somma nulla"); return;
        }

        int amount = 0;

        try {
            amount = Integer.parseInt(number);

            transaction = account.startGetMoney(amount);

            state = GET_MONEY_STARTED;

            logDebug("Sistema entrato nello stato GET_MONEY_STATE");

            gui.provideMoney(amount);

            gui.ejectCard(); card = null;

            logInfo("Fornita la somma " + amount + " ed espulsa la carta");
        } catch(NumberFormatException nfe) {
            error("La somma introdotta non e' un numero"); return;
        } catch(InvalidTransactionException ite) {
            error("La banca non ha autorizzato il prelievo"); return;
        } 
    }

    /**
     * E' stata completata l'operazione di prelievo.
     */
    public void getMoneyCompleted() {
        if(state != GET_MONEY_STARTED) { error("Errore interno del sistema"); return; }

        transaction.completeGetMoney();

        logInfo("Prelavata la somma di: " + transaction.getAmount());

        reset();
    }

    /**
     * E' stata annullata l'operazione di prelievo.
     */
    public void getMoneyCancelled() {
        if(state != GET_MONEY_STARTED) { error("Errore interno del sistema"); return; }

        transaction.rollbackGetMoney();

        logInfo("Transazione annullata");
   
        reset();
    }

    /**
     * L'utente non ha risposto entro un timeout fissato.
     */
    public void userTimeoutExpired() {
        if(transaction != null) transaction.rollbackGetMoney();

        if(card != null) { gui.retireCard(); card = null; }

        logInfo("L'utente non ha risposto entro il tempo massimo");

        reset();
    }

    /**
     * L'utente ha richiesto l'uscita forzata dal sistema.
     */
    public void forcedExit() {
        if(transaction != null) transaction.rollbackGetMoney();

        if(card != null) { gui.ejectCard(); card = null; }

        logInfo("L'utente ha richiesto l'uscita forzata");

        reset();
    }

    /**
     * Il sistema si riporta nello stato iniziale.
     */
    public void reset() {
        if(transaction != null && transaction.isOpen()) transaction.rollbackGetMoney();

        if(card != null) gui.ejectCard();

        logDebug("Sistema entrato nello stato INITIAL_STATE");

        state       = INITIAL_STATE; 
        transaction = null;
        card        = null;
        account     = null;
        pin         = null;
        user        = null;

        gui.showInitialScreen();
    }

    /**
     * Metodo utilizzato per indicare che si e' verificato un errore.
     * Riporta il sistema nello stato iniziale e visualizza l'errore.
     */
    public void error(String message) {
        state = ERROR_STATE;

        logDebug("Sistema entrato nello stato ERROR_STATE");

        logError(message);

        gui.showErrorScreen(message);
    }

    /**
     * Metodo che consente di scrivere informazioni sul file di log.
     */
    public void logInfo(String message) { log(INFO, message); }

    /**
     * Metodo che consente di scrivere errori sul file di log.
     */
    public void logError(String message) { log(ERROR, message); }

    /**
     * Metodo che consente di scrivere errori sul file di log.
     */
    public void logDebug(String message) { log(DEBUG, message); }

    /**
     * Tipi di messaggi possibili sul file di log.
     */
    private static final int INFO = 0, ERROR = 1, DEBUG = 2;

    /**
     * Metodo che consente di scrivere sul file di log.
     */
    private void log(int type, String message) {
        if(type == DEBUG && !test) return;

        if(logFile == null) return;

        switch(type) {
        case INFO:  logFile.println("[INFO ] " + message); break;
        case ERROR: logFile.println("[ERROR] " + message); break;
        case DEBUG: logFile.println("[DEBUG] " + message); break;
        }
    }

    /**
     * Stati in cui si pu trovare il controller.
     */
    private static final int INITIAL_STATE = 0, ERROR_STATE = 1, REQUEST_PIN_STATE = 2, VALIDATE_ACCOUNT_STATE = 3, MAIN_MENU_STATE = 4, ENTER_AMOUNT_STATE = 5, GET_MONEY_STARTED = 6;

    /**
     * La carta attualmente inserita il sistema.
     */
    private Card card;

    /**
     * Il conto corrente utilizzando correntemente dal sistema.
     */
    private Account account = null;

    /**
     * Il PIN utilizzando correntemente dal sistema.
     */
    private PIN pin = null;

    /**
     * L'utente che attualmente sta utilizzando il sistema.
     */
    private User user = null;

    /**
     * La transazione corrente.
     */
    private Transaction transaction = null;

    /**
     * Indica lo stato attuale del controller.
     */
    private int state = INITIAL_STATE;

    /**
     * Indica se il sistema e' lanciato in modalita' test.
     */
    private boolean test = false;

    /**
     * Oggetto che serve come interfaccia verso la banca.
     */
    private Bank bank = null;

    /**
     * Oggetto che serve da interfaccia verso l'utente.
     */
    private GUI gui = null;

    /**
     * Oggetto che consente di scrivere sul file di log.
     */
    private PrintStream logFile = null;
}
